#include "polyhedron.h"
#include "randomizer.h" 
#include <gl\gl.h>
#include <gl\glu.h>
#include <gl\glut.h>
#include <assert.h>

//Define switches===========================
//#define PLANETARY
	/* Arranges the composite assemblies so there is a big one in the middle
	with smaller ones around it. This is confusing. */
#define ASSEMBLYRADIUS 2.0
#define ASSEMBLYSCALE 0.7
	/* I use the ASSEMBLYRADIUS and ASSEMBLYSCALE parameters in
	the cAssemblyComposite::addRandomChildren method. */
#define MAXCHILDREN 3
//******** CONST STATICS **********************************************
const int cAssembly::WIREFRAMEFLAG = 1;
const int cAssembly::SHOWCONNECTORFLAG = 2;
const int cAssembly::SHOWRADIUSFLAG = 4;
int cAssembly::MAXDEPTH = 4;

const int cPolyhedron::TETRAHEDRON = 0;
const int cPolyhedron::CUBE = 1;
const int cPolyhedron::OCTAHEDRON = 2;
const int cPolyhedron::DODECAHEDRON = 3;
const int cPolyhedron::ICOSOHEDRON = 4;
const int cPolyhedron::SPHERE = 5;
const int cPolyhedron::CYLINDER = 6;
const int cPolyhedron::CONE = 7;
const int cPolyhedron::TORUS = 8;
const int cPolyhedron::TEAPOT = 9;


//Helper methods======================================== 
/* glutWireCylinder isn't defined, so we define it. */

GLUquadricObj *pquadObj = gluNewQuadric();

void glutWireCylinder(double bottomradius, double topradius, double height,
	int slices, int stacks)
{
	gluQuadricDrawStyle(pquadObj, GLU_LINE);
	gluQuadricNormals(pquadObj, GLU_SMOOTH);
	gluCylinder(pquadObj, bottomradius, topradius, height, slices, stacks);
}

void glutSolidCylinder(double bottomradius, double topradius, double height,
	int slices, int stacks)
{
	gluQuadricDrawStyle(pquadObj, GLU_FILL);
	gluQuadricNormals(pquadObj, GLU_SMOOTH);
	gluCylinder(pquadObj, bottomradius, topradius, height, slices, stacks);
}


void setVertexColor(cVector3 color, Real alpha)
{
  	float r = (float)color.x();
	float g = (float)color.y();
	float b = (float)color.z();
	CLAMP(r, 0.0, 1.0); //Use a macro from realnumber.h
	CLAMP(g, 0.0, 1.0);
	CLAMP(b, 0.0, 1.0);
	::glColor4f(r,g,b,alpha);
}

void setMaterialColor(cVector3 color)
{
  	float r = (float)color.x();
	float g = (float)color.y();
	float b = (float)color.z();
	CLAMP(r, 0.0, 1.0); //Use a macro from realnumber.h
	CLAMP(g, 0.0, 1.0);
	CLAMP(b, 0.0, 1.0);
	GLfloat materialAmbDiff[4]={ r, g, b, 1.0f };
	::glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, materialAmbDiff );
}

void setMaterialColor(Real r, Real g, Real b, Real alpha)
{
	CLAMP(r, 0.0, 1.0); //Use a macro from realnumber.h
	CLAMP(g, 0.0, 1.0);
	CLAMP(b, 0.0, 1.0);
	GLfloat materialAmbDiff[4]={ r, g, b, alpha};
	::glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, materialAmbDiff );
}


//=============================cAssembly=========================
//================================================================
//========Boring C++ Junk for cAssembly==============
cAssembly::cAssembly():
		_position(0,0,0),
		_actualradius(1.0), //Even though there's nothing there
		_scale(1.0, 1.0, 1.0),
		_cumulativeanimatespinangle(0.0), _animatespinaxis(0.0, 0.0, 1.0),
		_animatespinrate(0.0),
		_velocity(0.0, 0.0, 0.0)
{
	clearRotation(); // sets _rotationmatrix to zero degrees around z axis	
}

cAssembly::cAssembly(const cAssembly &assembly)
{
	*this = assembly; //The default operator = copies the fields ok.
}

cAssembly::cAssembly(cAssembly *passembly)
{
	*this = *passembly; //The default operator = copies the fields ok.
}

void cAssembly::copy(const cAssembly &assembly)
{
	*this = assembly; //The default operator = copies the fields ok.
}

void cAssembly::multiplyScale(cVector newscale)
{
	_scale = cVector(_scale.x()*newscale.x(), _scale.y()*newscale.y(),
		_scale.z()*newscale.z());
}

void cAssembly::multiplyScale(Real s)
{
	cVector sv(s, s, s);
	multiplyScale(sv);
}

//========Good Stuff for cAssembly====================
void cAssembly::randomize()
{//Randomize the animation parameters
	_animatespinaxis = cVector::randomUnitVector(); 
	_animatespinrate = cRandomizer::pinst()->randomReal(20.0, 60.0);
		/* The animationspinrate is measured in degrees per second. */
//Randomize the rotation
	cVector rotaxis = cVector::randomUnitVector();
	Real rotdegrees = cRandomizer::pinst()->randomReal(0.0, 360.0);
	_rotationmatrix = cMatrix::rotation(
		cSpin(ANGLETORADIAN*rotdegrees, rotaxis));
//Normally we won't randomize the scale or the position.
}

cAssembly* cAssembly::makeRandomAssembly(int depth)
{
	cAssembly *pnewassembly;
	if (depth >= 1)
	{
		pnewassembly = new cAssemblyComposite();
		((cAssemblyComposite*)pnewassembly)-> //Have to cast so I can use method
			addRandomChildren(depth); //Allow  maxdepth more levels.
	}
	else
		pnewassembly = new cPolyhedron();
	pnewassembly->randomize(); /* Randomizes the animationspinaxis and 
		animationspinrate, also,if it's a polyhedron, randomizes the color
		and shape. */
	pnewassembly->fixActualRadius();
	return pnewassembly;
}	

void cAssembly::setRotation(Real rdegrees, Real rx, Real ry, Real rz)
{ /* Sets_rotationmatrix to be a rotation of rdegrees around the axis
	(rx, ry, rz)*/
	cVector rotaxis(rx, ry, rz);
	cMatrix _rotationmatrix = cMatrix::rotation(
		cSpin(ANGLETORADIAN*rdegrees, rotaxis));
}

void cAssembly::applyRotation(Real rdegrees, Real rx, Real ry, Real rz)
{ /* Right-multiplies the current_rotationmatrix by a rotation of
	rdegrees around the axis (rx, ry, rz).  The applied rotation
	We wrap this change of rotation inside	a matrix called drotation. 
	Use the vectormatrix.h classes cMatrix and cSpin.  cSpin is based
	on radians, so we multiply by the ANGLETORADIAN
	conversion factor of PI/90.0 defined in realnumber.h */
	cVector rotaxis = cVector(rx, ry, rz);
	cMatrix drotation = cMatrix::rotation(
		cSpin(ANGLETORADIAN*rdegrees, rotaxis));
/* We apply drotation to our current _rotationmatrix by multiplying it on the
	right, which means it acts on the vertex BEFORE the old _rotationmatrix. 
	This way any initial positional adjustments of the assembly's rotation 
	still move the assembly, but AFTER these new twirls preserved. */
	_rotationmatrix = _rotationmatrix * drotation; //RIGHT multiply
//	_rotationmatrix = _rotationmatrix * drotation; //or try LEFT multiply
}

void cAssembly::animate(Real dt)
{
//We can update the position if we like.
	_position += dt*_velocity;
//Cacluate your desired change of spinangle as dspinangle
//You might possibly change your animationspinaxis her as well.
	Real dspinangle = dt*_animatespinrate;
/* Now we do the  rotation change, which is to right multiply by a rotation
 of dspinangle around the _animatespinaxis. */
	applyRotation(dspinangle, _animatespinaxis);
/* Track the added rotation in the _cumulativeanimatespinangle field, possibly for
clamping or reversing effects. Note however that we did NOT use _cumulativeanimatespinangle
in building our rotation. _cumulativeanimatespinangle is a dummy variable. */
	_cumulativeanimatespinangle += dspinangle; 
}
//=============================cAssemblyComposite=========================
//================================================================
//========Boring C++ Junk for cAssemblyComposite==============
cAssemblyComposite::~cAssemblyComposite()
{
	for (int i=0; i<_pchild.size(); i++)
		delete _pchild[i];
	_pchild.empty();	
}
cAssemblyComposite::cAssemblyComposite(const cAssemblyComposite &comp)
{
	copy(comp);
}
cAssemblyComposite::cAssemblyComposite(cAssemblyComposite *pcomp)
{
	copy(*pcomp);
}
void cAssemblyComposite::copy(const cAssemblyComposite &comp)
{
	cAssembly *pa;
	if (this == &comp)
		return;
	cAssembly::copy(comp);
	for (int i=0; i<_pchild.size(); i++)
	{
		pa = new cAssembly(comp._pchild[i]);
		_pchild.push_back(pa);	
	}
}

cAssemblyComposite cAssemblyComposite::operator=(const cAssemblyComposite &comp)
{
	if (this == &comp)
		return *this;
	copy(comp);
	return *this;
}

	//===========Good cAssemblyComposite Stuff===============

void cAssemblyComposite::addRandomChildren(int maxdepth)
{
	cAssembly *pnew;
	int childcount = cRandomizer::pinst()->random(2, MAXCHILDREN);
	for (int i=0; i<childcount; i++)
	{
		pnew = cAssembly::makeRandomAssembly(maxdepth - 1);
#ifndef PLANETARY
		pnew->setPosition(ASSEMBLYRADIUS * cVector::randomUnitVector());
		pnew->setScale(ASSEMBLYSCALE);// I'll scale the later guys to be smaller.
#else // PLANETARY means make the first one big and in the center.  Confusing.
		if (i>0)
		// I'll move the later guys out to be on periphery of the first guy.
		{
			cAssembly* pboss = _pchild[0];
			Real howfar = cRandomizer::pinst()->randomReal(0.5, 1.3);
			pnew->setPosition(howfar * pboss->radius() *
				 cVector::randomUnitVector());
			// I'll scale the later guys to be smaller.
			Real shrinkfactor = cRandomizer::pinst()->randomReal(0.5, 0.8);
			pnew->multiplyScale(shrinkfactor);
		}
#endif // PLANETARY
		add(pnew);
	}
}

	void cAssemblyComposite::animate(Real dt)
	{
		cAssembly::animate(dt);
		for (int i=0; i<_pchild.size(); i++)
			_pchild[i]->animate(dt);
	}


	int cAssemblyComposite::depth()
	{
		int maxdepth = 0;
		for (int i=0; i<_pchild.size(); i++)
			if (maxdepth < _pchild[i]->depth())
				maxdepth = _pchild[i]->depth();
		return maxdepth + 1;
	}

void cAssemblyComposite::fixActualRadius()
{
	cVector centerofmass(0,0,0);
	Real testradius, maxradius = 0;
	int i;
	for (i=0; i<_pchild.size(); i++)
	{
		_pchild[i]->fixActualRadius();
			/* Fix the _actualradius values of the child, just in case. */
		testradius =
			_pchild[i]->position().magnitude() + _pchild[i]->radius();
		if (testradius > maxradius)
			maxradius = testradius;
	}
	_actualradius = maxradius;
	if (!_actualradius)
		_actualradius = 1.0;
}

void cAssemblyComposite::add(cAssembly *pa)
{
	_pchild.push_back(pa);
}

void cAssemblyComposite::draw(int drawflags)
{
	static Real rvalue[7] = {1.0, 1.0, 0.0, 0.0, 1.0, 0.0, 1.0};
	static Real gvalue[7] = {0.0, 0.0, 1.0, 0.0, 1.0, 1.0, 1.0};
	static Real bvalue[7] = {1.0, 0.0, 0.0, 1.0, 0.0, 1.0, 1.0};
	int colorindex = depth();
	if (colorindex > 6)
		colorindex = 6;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
/* We make the OpenGL calls in the reverse order of how we want them to 
happen.  That is, the last call will act first on the vertices of the
polyhedron.  What we want to do is draw the children, then scale them, then
rotate them as a whole, then translate them to this _position. */
	glTranslatef(_position.x(), _position.y(), _position.z());
	glMultMatrixf(_rotationmatrix.transpose().elements());
	glScalef(_scale.x(), _scale.y(), _scale.z());
	for (int i=0; i<_pchild.size(); i++)
	{
		_pchild[i]->draw(drawflags);
		if (drawflags & SHOWCONNECTORFLAG)
		{
			setMaterialColor(rvalue[colorindex], gvalue[colorindex],
				bvalue[colorindex], 1.0);
			glBegin(GL_LINES);
				glVertex3d(0.0, 0.0, 0.0);
				(_pchild[i]->position()).glVertex();
			glEnd();
		} //end of the SHOWCONNECTORFLAG case
	} //End loop on children.  Have drawn children and connectors.
	glPopMatrix();
	if (drawflags & SHOWRADIUSFLAG)
	{
		glPushMatrix();
		setMaterialColor(rvalue[colorindex], gvalue[colorindex],
			bvalue[colorindex], 0.5);
		glTranslatef(_position.x(), _position.y(), _position.z());
		glMultMatrixf(_rotationmatrix.transpose().elements()); /* Apply
			the rotations accumulated in our cMatrix _rotationmatrix. */
		glutWireSphere(1.1*radius(), 8, 8);
		glPopMatrix();
	} //end SHOWRADIUSFLAG
}

//------------Constructors, Destructor, and Copiers---------------
//========Boring C++ Junk for cPolygon==============

cPolyhedron::cPolyhedron(int gluttype, Real radius,
	int slices, int stacks, Real edge,
	Real tuberadius, Real bottomradius,
	Real topradius, Real height):
		_gluttype(gluttype),
		_radius(radius),
		_slices(slices), _stacks(stacks), _edge(edge),
		_tuberadius(tuberadius), _bottomradius(bottomradius),
		_topradius(topradius), _height(height),
		_color(cVector(1.0, 0.0, 0.0)), _alpha(1.0),
		_filled(true), _linewidth(2.0)
{
	fixActualRadius();
	if (_gluttype == DODECAHEDRON)
		setScale(1.0/_actualradius); /* I do this correction here, and not
			in fixActualRadius, as fixActualRadius may get called more
			than once. In here, I know I only do it once, at start up. 
			The only reason I do this is because the dumb glut method makes
			all the other platonic solids of radius 1 by default but makes
			dodecahedron bigger, and I'd like them to start out of similar
			size. */
} 

void cPolyhedron::fixActualRadius()
{
	Real  alpha; //Need alpha and beta for DODECAHEDRON
	Real beta;
	Real bigradius; //Need bigradius for CYLINDER
	switch(_gluttype)
	{
		case TETRAHEDRON:
		case OCTAHEDRON:
		case ICOSOHEDRON:
			_actualradius =  1.0;
			break;
		case CUBE:
			_actualradius = (sqrt(0.75) * _edge); //Use 3D pythagorean theorem.
			break;
		case DODECAHEDRON:
			alpha = sqrt(2.0 / (3.0 + sqrt(5.0)));
			beta = 1.0 + sqrt(6.0 / (3.0 + sqrt(5.0)) -
				2.0 + 2.0 * sqrt(2.0 / (3.0 + sqrt(5.0))));
			_actualradius = sqrt(alpha*alpha + beta*beta);
			break;
		case SPHERE:
		case TEAPOT:
			_actualradius = _radius;
			break;
		case CYLINDER:
			bigradius = __max( _bottomradius, _topradius);
			_actualradius = sqrt(_height*_height + bigradius * bigradius);
			break;
		case CONE:
			_actualradius = __max(_bottomradius, _height);
			break;
		case TORUS:
			_actualradius = _radius + _tuberadius;
			break;
	}
}

cPolyhedron::cPolyhedron(const cPolyhedron &poly):
	cAssembly(poly)
{
	*this = poly;
}

//============Good Stuff for cPolyhedron=================

void cPolyhedron::randomize()
{
	cAssembly::randomize();
	_gluttype = cRandomizer::pinst()->random(TETRAHEDRON,TEAPOT);
		//Randomizer includes low and hi values of the argumetns. 
//	_gluttype = TORUS; //Can set if I just want to see one kind.
	fixActualRadius(); 
	_color = cRandomizer::pinst()->randomColor(0.5);  //Arg is brightness.
//	_filled = cRandomizer::pinst()->randomBool(0.9);
	/* As of molecules 4, I prefer not to have random unfilled
	polyhedra, as they are rendered so slowly. */
}

void cPolyhedron::draw(int drawflags)
{
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
/* We make the OpenGL calls in the reverse order of how we want them to 
happen.  That is, the last call will act first on the vertices of the
polyhedron.  What we want to do is draw the polyhedron,  then scale it,
 then rotate it, then translate it to its _position.  */
	glTranslatef(_position.x(), _position.y(), _position.z());
	glMultMatrixf(_rotationmatrix.transpose().elements());
		/* Apply the rotations accumulated in our cMatrix _rotationmatrix. */
	glScalef(_scale.x(), _scale.y(), _scale.z());
	setMaterialColor(_color);
	if (filled() & !(drawflags & WIREFRAMEFLAG))
	{
		switch(_gluttype)
		{
			case TETRAHEDRON:
				glutSolidTetrahedron();
				break;
			case CUBE:
				glutSolidCube(_edge);
					//argument is value of cube edge.
				break;
			case OCTAHEDRON:
				glutSolidOctahedron();
				break;
			case DODECAHEDRON:
				glutSolidDodecahedron();
				break;
			case ICOSOHEDRON:
				glutSolidIcosahedron();
				break;
			case SPHERE:
				glutSolidSphere(_radius, _slices, _stacks);
				break;
			case CYLINDER:
				glutSolidCylinder(_bottomradius, _topradius, _height,
					_slices, _stacks);
				break;
			case CONE:
				glutSolidCone(_bottomradius, _height, _slices, _stacks);
				break;
			case TORUS:
				glutSolidTorus(_tuberadius, _radius,
					_slices, _stacks);
				break;
			case TEAPOT:
				glutSolidTeapot(_radius);
				break;
		}
	}
	else //!filled() or you pass in a non-zero drawflag 
	{
		glLineWidth(_linewidth);
		switch(_gluttype)
		{
			case TETRAHEDRON:
				glutWireTetrahedron();
				break;
			case CUBE:
				glutWireCube(_edge);
					//argument is value of cube edge.
				break;
			case OCTAHEDRON:
				glutWireOctahedron();
				break;
			case DODECAHEDRON:
				glutWireDodecahedron();
				break;
			case ICOSOHEDRON:
				glutWireIcosahedron();
				break;
			case SPHERE:
				glutWireSphere(_radius, _slices, _stacks);
				break;
			case CYLINDER:
				glutWireCylinder(_bottomradius, _topradius, _height,
					_slices, _stacks);
				break;
			case CONE:
				glutWireCone(_bottomradius, _height, _slices, _stacks);
				break;
			case TORUS:
				glutWireTorus(_tuberadius, _radius,
					_slices, _stacks);
				break;
			case TEAPOT:
				glutWireTeapot(_radius);
				break;
		}
	}
	::glPopMatrix();
	if (drawflags & SHOWRADIUSFLAG)
	{
		::glPushMatrix();
		setMaterialColor(1.f, 0.0f, 1.0f, 0.5f);
		glTranslatef(_position.x(), _position.y(), _position.z());
		glMultMatrixf(_rotationmatrix.transpose().elements());
		glutWireSphere(1.1*radius(), 6, 6);
		::glPopMatrix();
	} //end SHOWRADIUSFLAG
}

